""" Code generation for explicitly specified tensors """

from dune.codegen.generation import (get_counted_variable,
                                     domain,
                                     kernel_cached,
                                     iname,
                                     instruction,
                                     temporary_variable,
                                     )

import pymbolic.primitives as prim
import numpy as np
import loopy as lp
import itertools as it


def define_assembled_tensor(name, expr, visitor):
    temporary_variable(name,
                       shape=expr.ufl_shape,
                       shape_impl=('fm',))
    for indices in it.product(*tuple(range(i) for i in expr.ufl_shape)):
        visitor.indices = indices
        instruction(assignee=prim.Subscript(prim.Variable(name), indices),
                    expression=visitor.call(expr),
                    forced_iname_deps=frozenset(visitor.interface.quadrature_inames()),
                    depends_on=frozenset({lp.match.Tagged("sumfact_stage1")}),
                    tags=frozenset({"quad"}),
                    )


@kernel_cached
def name_assembled_tensor(o, visitor):
    name = get_counted_variable("assembled_tensor")
    define_assembled_tensor(name, o, visitor)
    return name


@kernel_cached
def pymbolic_matrix_inverse(o, visitor):
    indices = visitor.indices
    visitor.indices = None
    name = name_assembled_tensor(o.ufl_operands[0], visitor)

    instruction(code="{}.invert();".format(name),
                within_inames=frozenset(visitor.interface.quadrature_inames()),
                depends_on=frozenset({lp.match.Writes(name),
                                      lp.match.Tagged("sumfact_stage1"),
                                      }),
                tags=frozenset({"quad"}),
                )

    visitor.indices = indices
    return prim.Variable(name)
